import logging
logger = logging.getLogger(__name__)


def limit_chord_unlock_tasks(app):
    """
    Set max_retries for chord.unlock tasks to avoid infinitely looping
    tasks. (see celery/celery#1700 or celery/celery#2725)
    """
    task = app.tasks['celery.chord_unlock']
    if task.max_retries is None:
        retries = getattr(app.conf, 'CHORD_UNLOCK_MAX_RETRIES', None)
        task.max_retries = retries


def setup_exchanges(app):
    """
    Setup result exchange to route all tasks to platform queue.
    """
    with app.producer_or_acquire() as P:
        # Ensure all queues are noticed and configured with their
        # appropriate exchange.
        for q in app.amqp.queues.values():
            P.maybe_declare(q)


SETUP_FUNCS = (
    limit_chord_unlock_tasks,
    setup_exchanges,
)


def setup_app(app, throw=True):
    """
    Ensure application is set up to expected configuration. This function is
    typically triggered by the worker_init signal, however it must be called
    manually by codebases that are run only as task producers or from within
    a Python shell.
    """
    success = True
    try:
        for func in SETUP_FUNCS:
            try:
                func(app)
            except:
                success = False
                if throw:
                    raise
                else:
                    msg = "Failed to run setup function %r(app)"
                    logger.exception(msg, func.__name__)
    finally:
        setattr(app, 'is_set_up', success)



from celery.utils.log import ColorFormatter as ColorFormatterBase


class ColorFormatter(ColorFormatterBase):
    def __init__(self, fmt, use_color=True, *args, **kwargs):
        super(ColorFormatter, self).__init__(fmt, use_color)



from os import environ as env
import pprint
import logging
import logging.config

from kombu import Exchange, Queue
from opbeat import Client
from opbeat.contrib.celery import register_signal
from opbeat.handlers.logging import OpbeatHandler

# Ensure signals are imported before app starts
from .signals import *  # NOQA
from . import DEFAULT_QUEUES


DEFAULT_LOGGING_FMT = '[%(asctime)s: %(levelname)s/%(processName)s %(message)s'
DEFAULT_LOGGING_CONFIG = {
    'version': 1,
    'formatters': {
        'colored': {
            'format': DEFAULT_LOGGING_FMT,
            'class': 'cadasta.workertoolbox.logging.ColorFormatter'
        },
        'default': {
            'format': DEFAULT_LOGGING_FMT,
        },
    },
    'handlers': {
        'console': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'colored',
        },
        'info_file': {
            'level': 'INFO',
            'class': 'logging.handlers.RotatingFileHandler',
            'formatter': 'default',
            'filename': 'app.info.log',
            'backupCount': 2,
            'maxBytes': 1024 * 1024 * 5,
        },
        'error_file': {
            'level': 'ERROR',
            'class': 'logging.handlers.RotatingFileHandler',
            'formatter': 'default',
            'filename': 'app.error.log',
            'backupCount': 2,
            'maxBytes': 1024 * 1024 * 5,
        },
    },
    'loggers': {
        '': {
            'handlers': ['console', 'info_file', 'error_file'],
        },
    }
}


class Config:
    # Broker
    QUEUE_NAME_PREFIX = env.get('QUEUE_PREFIX', 'dev')
    broker_transport = 'sqs'
    broker_transport_options = {
        'region': 'us-west-2',
        'queue_name_prefix': '{}-'.format(QUEUE_NAME_PREFIX)
    }

    # Results
    RESULT_DB_USER = env.get('RESULT_DB_USER', 'worker')
    RESULT_DB_PASS = env.get('RESULT_DB_PASS', 'cadasta')
    RESULT_DB_HOST = env.get('RESULT_DB_HOST', 'localhost')
    RESULT_DB_NAME = env.get('RESULT_DB_NAME', 'cadasta')
    RESULT_DB_PORT = env.get('RESULT_DB_PORT', '5432')
    result_backend = (
        'db+postgresql://{0.RESULT_DB_USER}:{0.RESULT_DB_PASS}@'
        '{0.RESULT_DB_HOST}:{0.RESULT_DB_PORT}/{0.RESULT_DB_NAME}')
    task_track_started = True

    # Exchanges
    task_default_exchange = 'task_exchange'
    task_default_exchange_type = 'topic'

    # Queues
    PLATFORM_QUEUE_NAME = 'platform.fifo'

    # Tasks
    CHORD_UNLOCK_MAX_RETRIES = 60 * 60 * 6  # 6hrs

    def __init__(self, QUEUES=DEFAULT_QUEUES, imports=('app.tasks',),
                 SETUP_LOGGING=True, **kw):
        """
        Object to manage Celery application configuration.
        """
        self.QUEUES = QUEUES
        self.imports = imports

        for k, v in kw.items():
            setattr(self, k, v)

        if SETUP_LOGGING:
            self.setup_default_logging()
            self.worker_hijack_root_logger = False

        try:
            self.result_backend = self.result_backend.format(self)
        except ValueError:
            raise ValueError(
                "Unable to render 'result_backend' value: %r" %
                self.result_backend)

        if not hasattr(self, 'task_queues'):
            self.task_queues = self._generate_queues(
                self.QUEUES, self._default_exchange_obj,
                self.PLATFORM_QUEUE_NAME)

        if not hasattr(self, 'task_routes'):
            self.task_routes = self._route_task

    def __repr__(self):
        attr_str = pprint.pformat(self.to_dict())
        return '{0.__class__.__name__}({1})'.format(self, attr_str)

    def to_dict(self):
        return {
            k: getattr(self, k)
            for k in dir(self)
            if (k.islower() and not k.startswith('_') and
                not callable(getattr(self, k)))
        }

    def setup_default_logging(self, opbeat_client=None):
        self.setup_log_files()
        if opbeat_client or env.get('OPBEAT_ORGANIZATION_ID'):
            self._opbeat_client = opbeat_client or Client()
            self.setup_opbeat_log_handler(self._opbeat_client)
            self.setup_opbeat_task_signal(self._opbeat_client)

    @staticmethod
    def setup_log_files(config=DEFAULT_LOGGING_CONFIG):
        logging.config.dictConfig(config)

    @staticmethod
    def setup_opbeat_log_handler(client, logger='', level=logging.ERROR):
        """
        Add OpBeat as log handler. Defaults to attaching to root logger
        and handling logs of level ERROR and above.
        """
        logger = logging.getLogger(logger)
        handler = OpbeatHandler(client)
        handler.setLevel(level)
        logger.addHandler(handler)

    @staticmethod
    def setup_opbeat_task_signal(client):
        """
        Setup OpBeat to handle Celery task failures.
        """
        register_signal(client)

    @property
    def _default_exchange_obj(self):
        return Exchange(
            self.task_default_exchange,
            self.task_default_exchange_type)

    @staticmethod
    def _generate_queues(queues, exchange, platform_queue):
        """ Queues known by this worker """
        return set([
            Queue('celery', exchange, routing_key='celery'),
            Queue(platform_queue, exchange, routing_key='#'),
        ] + [
            Queue(q_name, exchange, routing_key=q_name)
            for q_name in queues
        ])

    def _route_task(self, name, args, kwargs, options, task=None, **kw):
        return {
            'routing_key': name.split('.')[0],
            'exchange': self._default_exchange_obj
        }



import unittest
from unittest.mock import patch, MagicMock

from opbeat import Client

from cadasta.workertoolbox.conf import Config


class TestConfigClass(unittest.TestCase):

    def test_override(self):
        """ Ensure default params can be overridden """
        conf = Config(broker_transport='foo')
        self.assertEqual(conf.broker_transport, 'foo')

    def test_repr(self):
        conf = Config(a=1, b=2, z=3)
        self.assertTrue(repr(conf).startswith("Config({'a': 1,\n 'b': 2,\n"))
        self.assertTrue(repr(conf).endswith(" 'z': 3})"))

    def test_queues(self):
        self.assertEqual(
            Config(QUEUES=['foo']).QUEUES,
            ['foo'])
        self.assertEqual(
            Config(['bar']).QUEUES,
            ['bar'])

    def test_backend_default(self):
        self.assertEqual(
            Config(QUEUES=[]).result_backend,
            'db+postgresql://worker:cadasta@localhost:5432/cadasta')

    def test_backend_custom(self):
        self.assertEqual(
            Config(result_backend=':memory:').result_backend,
            ':memory:')

    def test_backend_custom_rendered(self):
        self.assertEqual(
            Config(
                result_backend='{0.foo}:{0.BAR}',
                foo='abc', BAR='def'
            ).result_backend,
            'abc:def'
        )

    def test_backend_custom_error(self):
        with self.assertRaises(ValueError) as context:
            Config(
                result_backend='{0.foo}:{}',
                foo='abc'
            )
        self.assertEqual(
            getattr(context.exception, 'message', context.exception.args[0]),
            "Unable to render 'result_backend' value: '{0.foo}:{}'"
        )

    def test_default_chord_unlock_max_retries(self):
        conf = Config()
        self.assertTrue(isinstance(conf.CHORD_UNLOCK_MAX_RETRIES, int))

    @patch('cadasta.workertoolbox.conf.env', {'OPBEAT_ORGANIZATION_ID': 123})
    @patch('cadasta.workertoolbox.conf.Client')
    @patch('cadasta.workertoolbox.conf.Config.setup_opbeat_log_handler')
    @patch('cadasta.workertoolbox.conf.Config.setup_opbeat_task_signal')
    def test_setup_opbeat_tools(self, task_signal, log_handler, Client):
        """ Ensure opbeat logging is called if env variable is set """
        Config()
        Client.assert_called_once_with()
        task_signal.assert_called_once_with(Client.return_value)
        log_handler.assert_called_once_with(Client.return_value)

    @patch('cadasta.workertoolbox.conf.env', {'OPBEAT_ORGANIZATION_ID': 123})
    @patch('cadasta.workertoolbox.conf.Client')
    @patch('cadasta.workertoolbox.conf.Config.setup_opbeat_log_handler')
    @patch('cadasta.workertoolbox.conf.Config.setup_opbeat_task_signal')
    def test_setup_opbeat_tools_override(self, task_signal, log_handler, Client):
        """
        Ensure opbeat logging is not called if env variable is set but setup
        is set to false
        """
        Config(SETUP_LOGGING=False)
        self.assertFalse(Client.called)
        self.assertFalse(task_signal.called)
        self.assertFalse(log_handler.called)

    @patch('cadasta.workertoolbox.conf.env', {})
    @patch('cadasta.workertoolbox.conf.Client')
    @patch('cadasta.workertoolbox.conf.Config.setup_opbeat_log_handler')
    @patch('cadasta.workertoolbox.conf.Config.setup_opbeat_task_signal')
    def test_setup_opbeat_tools_not_called(self, task_signal, log_handler, Client):
        """ Ensure opbeat logging is called if env variable is set """
        Config()
        self.assertFalse(Client.called)
        self.assertFalse(task_signal.called)
        self.assertFalse(log_handler.called)

    @patch('cadasta.workertoolbox.conf.logging')
    def test_setup_opbeat_log_handler(self, logging):
        """ Ensure opbeat logging handler adds handler to base logger """
        client = Client()
        Config.setup_opbeat_log_handler(client)
        logging.getLogger.assert_called_once_with('')
        logger = logging.getLogger.return_value
        self.assertEqual(logger.addHandler.call_count, 1)
        self.assertEqual(
            logger.addHandler.call_args_list[0][0][0].__class__.__name__,
            'OpbeatHandler')

    @patch('cadasta.workertoolbox.conf.logging')
    def test_setup_opbeat_log_handler_custom(self, logging):
        """ Ensure opbeat logging handler adds handler to custom logger """
        client = Client()
        Config.setup_opbeat_log_handler(client, 'foo')
        logging.getLogger.assert_called_once_with('foo')
        logger = logging.getLogger.return_value
        self.assertEqual(logger.addHandler.call_count, 1)
        self.assertEqual(
            logger.addHandler.call_args_list[0][0][0].__class__.__name__,
            'OpbeatHandler')

    @patch('cadasta.workertoolbox.conf.register_signal')
    def test_setup_opbeat_task_signal(self, register_signal):
        """ Ensure opbeat task handler calls opbeat signal setup function """
        client = Client()
        Config.setup_opbeat_task_signal(client)
        register_signal.assert_called_once_with(client)
